@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016 Nicira, Inc.
+ * Copyright (c) 2008-2016, 2019 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -220,9 +220,19 @@ want_to_poll_events(int want)
     }
 }
 
-/* Takes ownership of 'name'. */
+/* Creates a new SSL connection based on socket 'fd', as either a client or a
+ * server according to 'type', initially in 'state'.  On success, returns 0 and
+ * stores the new stream in '*streamp', otherwise returns an errno value and
+ * doesn't bother with '*streamp'.
+ *
+ * Takes ownership of 'name', which should be the name of the connection in the
+ * format that would be used to connect to it, e.g. "ssl:1.2.3.4:5".
+ *
+ * For client connections, 'server_name' should be the host name of the server
+ * being connected to, for use with SSL SNI (server name indication).  Takes
+ * ownership of 'server_name'. */
 static int
-new_ssl_stream(char *name, int fd, enum session_type type,
+new_ssl_stream(char *name, char *server_name, int fd, enum session_type type,
                enum ssl_state state, struct stream **streamp)
 {
     struct ssl_stream *sslv;
@@ -274,6 +284,14 @@ new_ssl_stream(char *name, int fd, enum session_type type,
     if (!verify_peer_cert || (bootstrap_ca_cert && type == CLIENT)) {
         SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);
     }
+#if OPENSSL_SUPPORTS_SNI
+    if (server_name && !SSL_set_tlsext_host_name(ssl, server_name)) {
+        VLOG_ERR("%s: failed to set server name indication (%s)",
+                 server_name, ERR_error_string(ERR_get_error(), NULL));
+        retval = ENOPROTOOPT;
+        goto error;
+    }
+#endif
 
     /* Create and return the ssl_stream. */
     sslv = xmalloc(sizeof *sslv);
@@ -293,6 +311,7 @@ new_ssl_stream(char *name, int fd, enum session_type type,
     }
 
     *streamp = &sslv->stream;
+    free(server_name);
     return 0;
 
 error:
@@ -301,6 +320,7 @@ error:
     }
     closesocket(fd);
     free(name);
+    free(server_name);
     return retval;
 }
 
@@ -311,6 +331,30 @@ ssl_stream_cast(struct stream *stream)
     return CONTAINER_OF(stream, struct ssl_stream, stream);
 }
 
+/* Extracts and returns the server name from 'suffix'.  The caller must
+ * eventually free it.
+ *
+ * Returns NULL if there is no server name, and particularly if it is an IP
+ * address rather than a host name, since RFC 3546 is explicit that IP
+ * addresses are unsuitable as server name indication (SNI). */
+static char *
+get_server_name(const char *suffix_)
+{
+    char *suffix = xstrdup(suffix_);
+
+    char *host, *port;
+    inet_parse_host_port_tokens(suffix, &host, &port);
+
+    ovs_be32 ipv4;
+    struct in6_addr ipv6;
+    char *server_name = (ip_parse(host, &ipv4) || ipv6_parse(host, &ipv6)
+                         ? NULL : xstrdup(host));
+
+    free(suffix);
+
+    return server_name;
+}
+
 static int
 ssl_open(const char *name, char *suffix, struct stream **streamp, uint8_t dscp)
 {
@@ -325,7 +369,8 @@ ssl_open(const char *name, char *suffix, struct stream **streamp, uint8_t dscp)
                              dscp);
     if (fd >= 0) {
         int state = error ? STATE_TCP_CONNECTING : STATE_SSL_CONNECTING;
-        return new_ssl_stream(xstrdup(name), fd, CLIENT, state, streamp);
+        return new_ssl_stream(xstrdup(name), get_server_name(suffix),
+                              fd, CLIENT, state, streamp);
     } else {
         VLOG_ERR("%s: connect: %s", name, ovs_strerror(error));
         return error;
@@ -514,6 +559,14 @@ ssl_connect(struct stream *stream)
             VLOG_INFO("rejecting SSL connection during bootstrap race window");
             return EPROTO;
         } else {
+#if OPENSSL_SUPPORTS_SNI
+            const char *servername = SSL_get_servername(
+                sslv->ssl, TLSEXT_NAMETYPE_host_name);
+            if (servername) {
+                VLOG_DBG("connection indicated server name %s", servername);
+            }
+#endif
+
             char *cn = get_peer_common_name(sslv);
 
             if (cn) {
@@ -899,7 +952,7 @@ pssl_accept(struct pstream *pstream, struct stream **new_streamp)
     ds_put_cstr(&name, "ssl:");
     ss_format_address(&ss, &name);
     ds_put_format(&name, ":%"PRIu16, ss_get_port(&ss));
-    return new_ssl_stream(ds_steal_cstr(&name), new_fd, SERVER,
+    return new_ssl_stream(ds_steal_cstr(&name), NULL, new_fd, SERVER,
                           STATE_SSL_CONNECTING, new_streamp);
 }
 
@@ -947,12 +1000,14 @@ do_ssl_init(void)
 {
     SSL_METHOD *method;
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined (LIBRESSL_VERSION_NUMBER)
 #ifdef _WIN32
     /* The following call is needed if we "#include <openssl/applink.c>". */
     CRYPTO_malloc_init();
 #endif
     SSL_library_init();
     SSL_load_error_strings();
+#endif
 
     if (!RAND_status()) {
         /* We occasionally see OpenSSL fail to seed its random number generator
@@ -1155,7 +1210,7 @@ stream_ssl_set_key_and_cert(const char *private_key_file,
                             const char *certificate_file)
 {
     if (update_ssl_config(&private_key, private_key_file)
-        || update_ssl_config(&certificate, certificate_file)) {
+        && update_ssl_config(&certificate, certificate_file)) {
         stream_ssl_set_certificate_file__(certificate_file);
         stream_ssl_set_private_key_file__(private_key_file);
     }
@@ -1252,12 +1307,12 @@ read_cert_file(const char *file_name, X509 ***certs, size_t *n_certs)
     }
 
     for (;;) {
-        X509 *certificate;
+        X509 *cert;
         int c;
 
         /* Read certificate from file. */
-        certificate = PEM_read_X509(file, NULL, NULL, NULL);
-        if (!certificate) {
+        cert = PEM_read_X509(file, NULL, NULL, NULL);
+        if (!cert) {
             size_t i;
 
             VLOG_ERR("PEM_read_X509 failed reading %s: %s",
@@ -1276,7 +1331,7 @@ read_cert_file(const char *file_name, X509 ***certs, size_t *n_certs)
         if (*n_certs >= allocated_certs) {
             *certs = x2nrealloc(*certs, &allocated_certs, sizeof **certs);
         }
-        (*certs)[(*n_certs)++] = certificate;
+        (*certs)[(*n_certs)++] = cert;
 
         /* Are there additional certificates in the file? */
         do {
